@@ -25,6 +25,7 @@
 #include <ogg/ogg.h>
 
 #include <gst/gst.h>
+#include <gst/base/gstflowcombiner.h>
 
 #include "gstoggstream.h"
 
@@ -88,6 +89,7 @@ struct _GstOggPad
 
   gint64 packetno;
   gint64 current_granule;
+  gint64 prev_granule;
   gint64 keyframe_granule;
 
   GstClockTime start_time;      /* the timestamp of the first sample */
@@ -95,7 +97,7 @@ struct _GstOggPad
   gint64 first_granule;         /* the granulepos of first page == first sample in next page */
   GstClockTime first_time;      /* the timestamp of the second page or granuletime of first page */
 
-  GstClockTime position;        /* position when last push occured; used to detect when we
+  GstClockTime position;        /* position when last push occurred; used to detect when we
                                  * need to send a newsegment update event for sparse streams */
 
   GList *continued;
@@ -127,6 +129,8 @@ struct _GstOggDemux
 
   GstPad *sinkpad;
 
+  GstFlowCombiner *flowcombiner;
+
   gint64 length;
   gint64 read_offset;
   gint64 offset;
@@ -144,6 +148,8 @@ struct _GstOggDemux
      useful for skewing when seeking */
   guint64 max_packet_size, max_page_size;
 
+  gboolean check_index_overflow;
+
   /* state */
   GMutex chain_lock;           /* we need the lock to protect the chains */
   GArray *chains;               /* list of chains we know */
@@ -157,7 +163,6 @@ struct _GstOggDemux
   GstSegment segment;
   guint32  seqnum;
 
-  GstEvent *event;
   GstEvent *newsegment;         /* pending newsegment to be sent from _loop */
 
   /* annodex stuff */
@@ -188,7 +193,6 @@ struct _GstOggDemux
   gboolean seek_secant;
   gboolean seek_undershot;
   GstClockTime push_prev_seek_time;
-  guint32 push_seek_seqnum;
 
   gint push_bisection_steps[2];
   gint stats_bisection_steps[2];
@@ -197,6 +201,17 @@ struct _GstOggDemux
 
   /* ogg stuff */
   ogg_sync_state sync;
+  long chunk_size;
+
+  /* Seek events set up by the streaming thread */
+  GstEvent *seek_event;
+  GThread *seek_event_thread;
+  GMutex seek_event_mutex;
+  GCond seek_event_cond;
+  gboolean seek_event_thread_stop;
+  gboolean seek_thread_started;
+  GCond thread_started_cond;
+  guint32 seek_event_drop_till;
 };
 
 struct _GstOggDemuxClass
@@ -204,8 +219,6 @@ struct _GstOggDemuxClass
   GstElementClass parent_class;
 };
 
-gboolean gst_ogg_demux_plugin_init (GstPlugin * plugin);
-
 G_END_DECLS
 
 #endif /* __GST_OGG_DEMUX_H__ */
